/*   COPYRIGHT International Business Machines Corp. 1997               */
/*   All Rights Reserved                                                */
/*   Licensed Materials - Property of IBM                               */
/*                                                                      */
/*   US Government Users Restricted Rights - Use, duplication or        */
/*   disclosure restricted by GSA ADP Schedule Contract with IBM Corp.  */

/**************************************************************/
/*                                                            */
/* Program Name:  BUILDCNV.C                                  */
/*                                                            */
/*  Program to generate a set of DFHCNV macros to match       */
/*   a COBOL copy book representing a CICS COMMAREA.          */
/*                                                            */
/*  Written by Jim MacNair - November, 1997                   */
/*                                                            */
/**************************************************************/

/**************************************************************/
/*                                                            */
/*  Current Version 1.03                                      */
/*                                                            */
/*  Changes in V1.01                                          */
/*                                                            */
/*   1) Corrected help display, adding space and new line.    */
/*                                                            */
/*  Changes in V1.02                                          */
/*                                                            */
/*   1) Removed DFHCNV TYPE=INITIAL and FINAL                 */
/*                                                            */
/*  Changes in V1.03                                          */
/*                                                            */
/*   1) Fixed bug in calculation of packed decimal lengths.   */
/*                                                            */
/*                                                            */
/**************************************************************/

/**************************************************************/
/*                                                            */
/*  This program handles simple COMMAREA definitions.  It     */
/*   does not support the DFHCNV selection capability for     */
/*   multiply defined areas.  The program will use            */
/*   the first definition for a particular area that it       */
/*   finds in the copy book.  Later REDEFINES are ignored.    */
/*                                                            */
/**************************************************************/

/**************************************************************/
/*                                                            */
/* This program takes a COBOL copy book as input.  The copy   */
/* book is assumes to represent a CICS COMMAREA.  The         */
/* program first builds an in memory data structure to        */
/* represent the copy book.  Each variable definition is      */
/* entered into a variable table.  By default, the table      */
/* can handle up to 4096 variables - this can be expanded     */
/* by changing the MAX_VAR_COUNT definition and recompiling   */
/* the program.                                               */
/*                                                            */
/* The following COBOL keywords are recognized and used as    */
/*  indicated.                                                */
/*                                                            */
/*  PIC - The picture clause following is processed.  If the  */
/*   keyword IS is found after PICTURE, it is skipped and     */
/*   ignored.  The picture clause is used to determine the    */
/*   length of variables, and to a lesser extent their type.  */
/*   The USAGE of a variable is also used to determine the    */
/*   length and type of a variable.                           */
/*                                                            */
/*  BINARY - Indicated that the usage is BINARY.  The         */
/*   length of the item is adjusted to 2, 4 or 8 bytes,       */
/*   depending on the picture clause.                         */
/*                                                            */
/*  COMP - Treated the same as BINARY.                        */
/*                                                            */
/*  COMP-1 and COMP-2 - Floating point, with COMP-1 being     */
/*   4 bytes long and COMP-2 being 8 bytes long.              */
/*                                                            */
/*  COMP-3 - Packed decimal - the length is adjusted based    */
/*   on the picture clause.                                   */
/*                                                            */
/*  COMP-4 and COMP-5 - treated the same as BINARY.           */
/*                                                            */
/*  POINTER and PROCEDURE-POINTER - Length is set to 4 bytes. */
/*                                                            */
/*  OCCURS - The number of occurences is saved, and used to   */
/*   calculate the offsets of items which follow.  OCCURS     */
/*   clauses can be given for higher level items as well      */
/*   as primitive data types, and can be nested.              */
/*                                                            */
/*  REDEFINES - The variable that is redefined is located     */
/*   and a pointer is established.  Redefines can be for      */
/*   higher level items as well as primitive data types,      */
/*   and can be nested.                                       */
/*                                                            */
/*  JUST - The type of justification is noted.                */
/*                                                            */
/*  SYNC - The alignment boundary is saved.  This can be      */
/*   a 2, 4 or 8 byte boundary, depending on the USAGE of     */
/*   the item.                                                */
/*                                                            */
/*  The following COBOL keywords are ignored.                 */
/*                                                            */
/*   USAGE - the actual usage types, such as COMP or BINARY   */
/*    are recognized and used.                                */
/*                                                            */
/*   IS - optional in COBOL, safely ignored.                  */
/*                                                            */
/*   DISPLAY - ignored, since this is the default.            */
/*                                                            */
/*   TIMES - optional and ignored.                            */
/*                                                            */
/*   BLANK WHEN ZERO.                                         */
/*                                                            */
/*   VALUE - should not be specified anyway.                  */
/*                                                            */
/*   EXTERNAL - should not be specified.                      */
/*                                                            */
/*   SIGN.                                                    */
/*                                                            */
/*   GLOBAL.                                                  */
/*                                                            */
/**************************************************************/

/**************************************************************/
/*                                                            */
/* Handling of SYNC clauses                                   */
/*                                                            */
/*  SYNC or SYNCHRONIZED clauses are only supported on        */
/*   primitive data items.  They will align the current       */
/*   variable on a storage address boundary, depending on     */
/*   the data type and length.  This applies to binary,       */
/*   index and pointer data types, and they will be aligned   */
/*   on a boundary which is an even multiple of their data    */
/*   length.  Slack bytes will be inserted as required to     */
/*   align the item.  The slack bytes are not part of any     */
/*   primitive data item, but will be included in group       */
/*   items.                                                   */
/*                                                            */
/*  The slack bytes are inserted before a variable.  They     */
/*   also affect the alignment and length of group variables. */
/*   If the first variable in a group variable is aligned,    */
/*   then the group item will also be aligned.                */
/*                                                            */
/*  The SYNC clause also interacts with an OCCURS clause.     */
/*   If an OCCURS clause is used on a group level, then       */
/*   slack bytes may be required both at the beginning of     */
/*   the item and before each occurence of the group          */
/*   variable.  The total length of the group item will be    */
/*   rounded up to a boundary that is a multiple of the       */
/*   largest alignment boundary found with the group item.    */
/*   These slack bytes will be added to the end of the        */
/*   group variable.                                          */
/*                                                            */
/**************************************************************/

/**************************************************************/
/* Required include for C functions                           */
/**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/**************************************************************/
/*                                                            */
/* Defines and global variables                               */
/*                                                            */
/**************************************************************/

#define MAX_LINE_LEN  255
#define MAX_STMT_LEN  2550

/**************************************************************/
/* Variable table size definitions                            */
/**************************************************************/

#define MAX_NAME_LEN   31
#define MAX_DATA_LEN   64
#define MAX_VAR_COUNT  4096

/**************************************************************/
/* Variable table structure definition                        */
/*                                                            */
/* The variable table includes all the characteristics of     */
/* the variable, as follows:                                  */
/*                                                            */
/*  VarName - Variable name, up to 31 characters.             */
/*                                                            */
/*  VarLevel - The COBOL level, from 01 to 49.  Note that     */
/*   level 66, 77 and 88 items are ignored.                   */
/*                                                            */
/*  VarOffset - The offset into the records where this        */
/*   variable begins.                                         */
/*                                                            */
/*  VarLen - The length of this item.  For primitive items,   */
/*   the length is calculated from the picture and/or usage   */
/*   clauses.  For higher levels, the length is calculated    */
/*   from the lengths of the individual items contained in    */
/*   the higher level variable.                               */
/*                                                            */
/*  VarParent - This is a pointer to the next higher level    */
/*   item.  For example, if an 03 level is defined under      */
/*   an 02 level, the VarParent of the 03 level would point   */
/*   to the 02 variable as its parent.                        */
/*                                                            */
/*  VarChild - Pointer to the first child.  This field is     */
/*   used as an indicator that this variable is not a         */
/*   primitive data area.                                     */
/*                                                            */
/*  VarLastChild - Pointer to the last child.  This field is  */
/*   used to find the last variable that is included in       */
/*   the definition of this variable.                         */
/*                                                            */
/*  VarNextLevel - A pointer to the next variable at the      */
/*   same level.  For example, all 02 level variables are     */
/*   chained together.  The anchor for this chain is the      */
/*   firstatlevel variable, and a chain ends with a zero      */
/*   pointer.                                                 */
/*                                                            */
/*  VarPrevLevel - A pointer to the previous variable at the  */
/*   same level.  The anchor for this chain is the            */
/*   nextatlevel variable, and a chain ends with a zero       */
/*   pointer.                                                 */
/*                                                            */
/*  VarRedefine - If a redefines clause was found, this is    */
/*   a pointer to the redefined variable.  If no redefines    */
/*   clause is found, then this pointer is zero.              */
/*                                                            */
/*  VarOccur - If an occurs clause was found for this         */
/*   variable, then the VarOccur will contain the number of   */
/*   times this variable occurs.  If no occurs clause is      */
/*   found, then this clause will contain a zero.             */
/*                                                            */
/*  VarSync - If a SYNCHRONIZED clause is found, then this    */
/*   field contains the alignment boundary for this field,    */
/*   which is determined by the USAGE clause.                 */
/*                                                            */
/*  VarMaxSync - If a SYNCHRONIZED clause is found within a   */
/*   group variable with an OCCURS clause, this field will be */
/*   used to remember the largest alignment of any individual */
/*   field.                                                   */
/*                                                            */
/*  VarEndSlack - If a SYNCHRONIZED clause is found within a  */
/*   group variable with an OCCURS clause, this field will    */
/*   hold the number of slack bytes at the end of the         */
/*   variable.                                                */
/*                                                            */
/*  VarType - The variable type is calculated based on the    */
/*   picture clause.  The type will be based on the first     */
/*   recognized character in the picture clause, such as      */
/*   an X or a 9.                                             */
/*                                                            */
/*  VarUsage - If a usage clause is found, then this          */
/*   entry is set to the particular usage, and the length     */
/*   of the data entry is adjusted to match the usage.        */
/*   The default usage is display.                            */
/*                                                            */
/*  VarJust - If a JUSTIFIED clause is found, then the        */
/*   type of justification is stored in this field,           */
/*   otherwise this field contains a binary zero.             */
/*                                                            */
/**************************************************************/

typedef struct {
   int  VarLevel;
   int  VarOffset;
   int  VarLen;
   int  VarParent;
   int  VarChild;
   int  VarLastChild;
   int  VarNextLevel;
   int  VarPrevLevel;
   int  VarRedefine;
   int  VarOccur;
   int  VarSync;
   int  VarMaxSync;
   int  VarEndSlack;
   char VarType;
   char VarUsage;
   char VarJust;
   char VarName[MAX_NAME_LEN + 1];
} VARSTRUCT;

/*****************************************************************/
/*                                                               */
/* COBOL Data Division keyword definitions                       */
/*                                                               */
/*****************************************************************/

#define PIC_STR       "PIC"
#define OCCURS_STR    "OCCURS"
#define SYNC_STR      "SYNC"
#define JUST_STR      "JUST"
#define COMP_STR      "COMP"
#define BINARY_STR    "BINARY"
#define PACKED_STR    "PACKED"
#define POINTER_STR   "POINTER"
#define PROCEDURE_STR "PROCEDURE"
#define DISPLAY_STR   "DISPLAY"
#define INDEX_STR     "INDEX"
#define REDEFINES_STR "REDEFINES"
#define USAGE_STR     "USAGE"
#define BLANK_STR     "BLANK"
#define VALUE_STR     "VALUE"
#define IS_STR        "IS"
#define FILLER_STR    "FILLER"

/*****************************************************************/
/*                                                               */
/*  Data type definitions - from USAGE clauses.                  */
/*                                                               */
/*****************************************************************/

#define USE_DISPLAY   0
#define USE_BINARY    1
#define USE_POINTER   2
#define USE_PACKED    3
#define USE_FLOAT1    4
#define USE_INDEX     5
#define USE_FLOAT2    8

/*****************************************************************/
/*                                                               */
/* defines for DFHCNV TYPE=FIELD statements.                     */
/* CHARACTER is character data, which must be translated from    */
/*  EBCDIC to ASCII.                                             */
/* BINARY is 370 binary data and is not changed.                 */
/* PD is packed decimal and is not changed.                      */
/* NUMERIC is data in Intel integer format and must be changed   */
/*  from little-endian to big-endian.                            */
/*                                                               */
/*****************************************************************/

#define CNV_CHARACTER  "CHARACTER"
#define CNV_PD         "PD"
#define CNV_BINARY     "BINARY"
#define CNV_NUMERIC    "NUMERIC"

static char copyright[] = "(C) Copyright IBM Corp, 1998, 2000";
static char Version[]=\
"@(#)buildcnv V1.03- DFHCNV macro creation utility  - Jim MacNair ("__DATE__")";

/*****************************************************************/
/*                                                               */
/* Global Debug Variable - If debugging information is desired,  */
/*  set this variable to 1 and recompile.                        */
/*                                                               */
/*****************************************************************/

DEBUG=0;

/**************************************************************/
/*                                                            */
/* This routine skips blanks in a string                      */
/*                                                            */
/**************************************************************/

char * skipBlanks(char *str)

{
   while (*str == ' ')
   {
      str++;
   } /* endwhile */

   return str;
}

/**************************************************************/
/*                                                            */
/* This routine finds the next blank in a string              */
/*                                                            */
/**************************************************************/

char * findBlank(char *str)

{
   while ((str[0] != ' ') && (str[0] != 0))
   {
      str++;
   } /* endwhile */

   return str;
}

/**************************************************************/
/*                                                            */
/* Routine to find a variable in the variable table           */
/*                                                            */
/**************************************************************/

int findVar(char *varname, VARSTRUCT *fvar)

{
   int i = 1;

   while ((i < MAX_VAR_COUNT) && (fvar[i].VarName[0] != 0) &&
          strcmp(fvar[i].VarName,varname))
    {
      i++;
    } /* endwhile */

   /* return the offset if found, otherwise return a zero */
   if (fvar[i].VarName[0] == 0)
   {
      return 0;
   }
   else
   {
      return i;
   } /* endif */
}

/**************************************************************/
/*                                                            */
/* This routine finds the value in a parameter within         */
/* parentheses.                                               */
/*                                                            */
/**************************************************************/

void getValue(char *map, char *mapName)

{
   char *beginVal;
   char *endVal;

   *(mapName) = '\0';  /* set parameter value to null */

   beginVal = strchr(map, '('); /* search for begin parentheses */
   if (beginVal != NULL)
   {
      beginVal++;  /* move past begin parentheses */
      endVal = strchr(beginVal, ')'); /* find end parentheses */
      if (endVal != NULL)
      {
         *endVal = '\0'; /* terminate the string */
         strcpy(mapName,beginVal);
      }
   }
}

/**************************************************************/
/*                                                            */
/* This routine writes a line to the message file             */
/* that is produced with the new program.                     */
/*                                                            */
/**************************************************************/

void writeErrorLine(int count, char *msg, FILE *msgfile)

{
   /* print the message */
   fprintf(msgfile,"line %.4d - %s\n",count + 1,msg);
}

/**************************************************************/
/*                                                            */
/* This routine finds the target variable name of a           */
/* redefines clause and moves it into                         */
/*                                                            */
/**************************************************************/

void processRedefine(char *redefines, char *redefineName)

{
   char *varptr;
   int  i=0;

   /* Initiatize variable name to null */
   redefineName[0] = 0;

   /* Initialize pointer */
   varptr = redefines;

   /* search for a delimiter after redefines */
   varptr = findBlank(redefines);

   if (varptr[0] != 0)
   {
      /* find the variable name that is redefined */
      varptr = skipBlanks(varptr);
      if (varptr[0] != 0)
      {
         /* Move the variable name to the designated area */
         while ((varptr[0] != ' ') && (varptr[0] != 0) && (i < MAX_NAME_LEN))
         {
            redefineName++[0] = varptr++[0]; /* Move the next byte */
            redefineName[0] = 0;   /* Terminate the string */
            i++;                   /* Increment length counter */
         } /* endwhile */
      } /* endif */
   } /* endif */
}

/**************************************************************/
/*                                                            */
/* This routine adjusts the length of the picture             */
/* clause to match a binary number's real length              */
/*                                                            */
/**************************************************************/

int handleBinaryLength(int piclength)

{
   int len;

   /* adjust the length based on the picture */
   if (piclength < 5)
   {
      /* handle pic 9(1) to 9(4) */
      len = 2;
   }
   else
   {
      /* handle pic 9(5) or greater */
      len = 4;
   }

   return len;
}

/**************************************************************/
/*                                                            */
/* This routine adjusts the length of a usage PACKED          */
/* clause mask element.  It runs after the picture clause,    */
/* if any, has been parsed.                                   */
/*                                                            */
/**************************************************************/

int handlePackedLength(int length)

{
   int  len=0;

   /* divide the length by 2 to account for packed, and allow for the sign byte */
   len = (length / 2) + 1;

   /* set the usage and the length */
   return len;
}

/**************************************************************/
/*                                                            */
/* This routine handles a usage COMP clause mask element      */
/* It runs after the picture clause, if any, has been         */
/* parsed.  It will adjust or set the length from the         */
/* picture clause as needed.                                  */
/*                                                            */
/**************************************************************/

char handleComp(char *usage, int *length)

{
   char *loc;
   char use;

   /* step past the keyword COMP */
   loc = usage + strlen(COMP_STR);

   /* set if just plain comp */
   use = USE_BINARY;

   /* next, check if there is a specific type of comp.   */
   /* note that the keyword can be COMP or COMPUTATIONAL */
   while ((loc[0] != ' ') && (loc[0] != '-') &&
          (loc[0] != '\0') && (loc[0] != '.'))
   {
       loc++;
   }

   /* check if we found a dash */
   if (loc[0] == '-')
   {
      switch (loc[1])
      {
         case '1':
            /* set the usage and the length */
            use = USE_FLOAT1;
            *length = 4;
            break;

         case '2':
            /* set the usage and the length */
            use = USE_FLOAT2;
            *length = 8;
            break;

         case '3':
            /* set the usage */
            use = USE_PACKED;

            /* adjust the length */
            *length = handlePackedLength(*length);

            break;

         case '4':
            /* set the usage */
            use = USE_BINARY;

            /* calculate the actual length of the item */
            (*length) = handleBinaryLength(*length);

            break;

         case '5':
            /* set the usage */
            use = USE_BINARY;

            /* calculate the actual length of the item */
            (*length) = handleBinaryLength(*length);

            break;

         default:
            break;

      }
   }
   else
   {
      /* no dash - must be plain COMP item       */
      /* calculate the actual length of the item */
      (*length) = handleBinaryLength(*length);
   }

   return use;
}

/**************************************************************/
/*                                                            */
/* This routine handles an OCCURS clause.                     */
/*                                                            */
/**************************************************************/

int handleOccurs(char *occurs)

{
   int  count=0;
   char *countend;

   /* find the delimiting blank */
   occurs = findBlank(occurs);

   /* Make sure we don't run past the end of the statement */
   if (occurs[0] != 0)
   {
      /* find the count */
      occurs = skipBlanks(occurs);

      /* find the end of the count and add a delimiter */
      countend = findBlank(occurs);
      countend[0] = 0;

      /* Convert the count to an integer */
      count = atoi(occurs);
   } /* endif */
   return count;
}

/*******************************************************************/
/*                                                                 */
/* This routine handles a picture clause mask element.  It sets    */
/*  the picture type as appropriate, and updates the picture       */
/*  length.  It handles counts following a picture element, as     */
/*  in X(9) or 9(4).  The location pointer within the picture is   */
/*  updated and returned to the calling routine.                   */
/*                                                                 */
/*******************************************************************/

char * handleElement(char *pic, char *type, int *len)

{
   char *loc;
   int  length=1;

   /* Set the type if it has not been previously set */
   if (type[0] == 0)
   {
      *type = *pic;
   }

   /* get location of the next character */
   loc = pic + 1;

   if (loc[0] == '(')
   {
      /* find the end of the count */
      while((loc[0] != ')') && (loc[0] != '\0'))
      {
         loc++;
      }

      /* move on past the close parentheses, if found */
      if (loc[0] == ')')
      {
         /* terminate the string */
         loc[0] = '\0';

         loc++;
      }

      /* get the count of the number of characters */
      length = atoi(pic + 2);
   }

   /* Check for DBCS character */
   if ((type[0] == 'G') || (type[0] == 'N'))
   {
      /* two bytes per character */
      length = length << 1;
   } /* endif */

   /* update the length */
   *len = (*len) +length;

   /* Return the new location pointer */
   return loc;
}

/*******************************************************************/
/*                                                                 */
/* This routine handles a picture clause.                          */
/*                                                                 */
/* The following characters in a picture clause occupy one         */
/*  position:                                                      */
/*                                                                 */
/*    A - Alphabetic Character                                     */
/*    B - Blank Character                                          */
/*    X - Alphameric Character                                     */
/*    9 - Numeric Character                                        */
/*    . - Period                                                   */
/*    , - Comma                                                    */
/*    * - Asterisk (Check protect character)                       */
/*    $ - Dollar Sign (Currency Symbol)                            */
/*    / - Slash, used as delimiter in dates                        */
/*                                                                 */
/*  The following characters occupy no space:                      */
/*                                                                 */
/*    E - Exponent in floating point number                        */
/*    P - Decimal point scaling position                           */
/*    S - Indicates signed number                                  */
/*    V - Decimal point assumed position                           */
/*                                                                 */
/*  The following characters or sequences occupy two positions:    */
/*                                                                 */
/*    CR - Indicates credit balance                                */
/*    DB - Indicates debit balance                                 */
/*    N  - DBCS Character                                          */
/*    G  - DBCS Character                                          */
/*                                                                 */
/*  This routine will attempt to determine the type of data        */
/*  within a field, for purposes of justification, translation,    */
/*  etc.                                                           */
/*                                                                 */
/*  This routine does not handle mixed type fields, to the extent  */
/*  they are legal within COBOL.  It will set the field type       */
/*  to match the first recognized type character, and will only    */
/*  set the type to A (Alphabetic), 9 (Numeric) or X (Alphameric). */
/*                                                                 */
/*******************************************************************/

int handlePic(char *pic, char *type)

{
   char *val;
   char savetype=0;
   int  length=0;

   /* point past the PIC keyword */
   val = pic + 3;

   /* find the end of the PIC keyword */
   while ((val[0] != ' ') && (val[0] != '\0'))
   {
      val++;
   }

   /* find the picture mask */
   val = skipBlanks(val);

   /* process the mask */
   while ((val[0] != ' ') && (val[0] != '\0'))
   {
      switch (val[0])
      {
         /* These characters take a space and set the variable   */
         /*  type.                                               */
         case 'A':      /* Alphabetic character */
         case 'X':      /* Alphameric character */
         case '9':      /* Numeric character    */
            val = handleElement(val, &savetype, &length);
            break;

         /* Numeric character, with leading zero suppression.    */
         /* Treat the same as a numeric character.               */
         case 'Z':
            val = handleElement("9", &savetype, &length);
            break;

         /* These characters take a space but do not set the     */
         /*  variable type.                                      */
         case '0':      /* Constant zero                 */
         case '/':      /* Constant slash, used in dates */
         case '.':      /* Constant period               */
         case ',':      /* Constant comma                */
         case '*':      /* Check protection character    */
         case '$':      /* Currency symbol               */
         case 'B':      /* Blank space                   */
         case 'C':      /* First character of CR         */
         case 'D':      /* First character of DB         */
         case 'R':      /* Second character of CR        */
         case 'N':      /* DBCS Character                */
         case 'G':      /* DBCS Character                */
            val = handleElement("\0", &savetype, &length);
            break;

         /* These characters do not take a space and are ignored */
         /*  as are any unrecognized characters.                 */
         case 'E':      /* Exponent delimiter             */
         case 'P':      /* Decimal precision              */
         case 'S':      /* Signed number                  */
         case 'V':      /* Assumed Decimal point location */
         default:
            val++;
            break;
      }
   }

   type[0] = savetype;

   /* return the length of the picture clause */
   return length;
}

/*******************************************************************/
/*                                                                 */
/*  Subroutine to write out an individual DFHCNV TYPE=FIELD        */
/*   entry.                                                        */
/*                                                                 */
/*******************************************************************/

void writeField(int offset, char use, int len, char *last, FILE *cnvfile)

{
   char      *type;
   char      cnvline[80];

   /* figure out the data type */
   switch (use)
   {
      case  USE_DISPLAY:
         type = CNV_CHARACTER;
         break;

      case  USE_BINARY:
         if (len > 1)
         {
            type=CNV_NUMERIC;
         }
         else
         {
            type=CNV_BINARY;
         } /* endif */

         break;

      case  USE_PACKED:
         type=CNV_PD;
         break;

      case  USE_POINTER:
      case  USE_FLOAT1:
      case  USE_INDEX:
      case  USE_FLOAT2:
         type = CNV_BINARY;
         break;

      default:
         break;
   } /* endswitch */

   sprintf(cnvline,"         DFHCNV TYPE=FIELD,OFFSET=%d,DATATYP=%s,",
           offset, type);
   fprintf(cnvfile,"%-71s*\n", cnvline);
   fprintf(cnvfile,"               DATALEN=%d%s\n",
           len,last);
}

/*******************************************************************/
/*                                                                 */
/*  Subroutine to handle OCCURS clauses when writing out the CNV   */
/*   macros.                                                       */
/*                                                                 */
/*******************************************************************/

void writeCNVOccurs(VARSTRUCT *var, int firstchild, int varcount, int occ,
                    char *last, FILE *cnvfile)

{
   int       offset;
   int       align;
   VARSTRUCT *curvar;
   int       i;
   char      *lst;

   /* point to the first variable */
   curvar = var + firstchild;

   /* get the beginning offset */
   offset = curvar->VarOffset;

   /* Handle OCCURS entries */
   while (occ > 0)
   {
      /* Handle all the variables */
      for (i = 0; i < varcount; i++)
      {
         /* Check if we have the last variable */
         if ((occ == 1) && (i == varcount - 1))
         {
            lst = last;
         }
         else
         {
            lst = "";
         }

         /* write out an individual field entry */
         writeField(offset, curvar->VarUsage, curvar->VarLen, lst, cnvfile);

         /* Increment the offset by the length */
         offset += curvar->VarLen;

         /* Check for any required alignment */
         if (curvar->VarSync > 1)
         {
            align = offset%curvar->VarSync;
            if (align > 0)
            {
               offset += curvar->VarSync - align;
            }
         }

                   curvar++;
      }

      /* point to the first variable */
      curvar = var + firstchild;

      /* Decrement the OCCURS counter */
      occ--;
   }
}

/*******************************************************************/
/*                                                                 */
/* Subroutine to write out a complete DFHCNV Macro file for use    */
/*  on a host CICS system.  First, the variable table is scanned   */
/*  and a new one is built, dropping any variables that are in     */
/*  REDEFINES clauses.  The next step is to write out the actual   */
/*  DFNCNV macros to a file.  Except for OCCURS clauses, each      */
/*  variable will generate one DFHCNV TYPE=FIELD entry.            */
/*                                                                 */
/*******************************************************************/

void buildCNV(FILE *cnvfile, VARSTRUCT *fvar, int maxvar)

{
   VARSTRUCT *var;
   VARSTRUCT *savevar;
   char      *last;
   int       redeflevel=0;
   int       maxcnv=0;
   int       i;
   int       j;
   int       occ;
   int       varcount;         /* Work variable for OCCURS clauses */
   int       varnum;

   /**************************************************************/
   /*                                                            */
   /*     Allocate and initialize variable areas                 */
   /*                                                            */
   /**************************************************************/

   var = (VARSTRUCT *)malloc(sizeof(VARSTRUCT) * (maxvar + 1));
   memset(var,'\0',(sizeof(VARSTRUCT) * (maxvar + 1)));

   /* initialize our pointer to the second entry */
   fvar++;

   /* remember the address of the new variable table */
   savevar = var;

   /**************************************************************/
   /*                                                            */
   /* This section will scan the full variable table, looking    */
   /* for variables which should generate an entry in the CNV    */
   /* table.  This includes only primitive variables and not     */
   /* higher level variables.  Variables in redefined areas are  */
   /* also ignored.                                              */
   /* Variables which have an OCCURS statement will be included  */
   /* as well, since they are required to process the OCCURS     */
   /* clauses.                                                   */
   /*                                                            */
   /**************************************************************/

   for (i=0; i < maxvar; i++)
   {
      /* Check if we have ended the current REDEFINES clause */
      if (redeflevel >= fvar->VarLevel)
      {
         redeflevel = 0;
      }

      /* Check if we are beginning a new REDEFINES clause */
      if ((fvar->VarRedefine > 0) && (fvar->VarLevel > redeflevel))
      {
         redeflevel = fvar->VarLevel;
      }

      /* Ignore if we are in a REDEFINES clause or if this is    */
      /* not a primitive variable.                               */
      if ((redeflevel == 0) && ((fvar->VarChild == 0) || (fvar->VarOccur > 0)))
      {
         memcpy(var, fvar, sizeof(VARSTRUCT));
         var++;
         maxcnv++;
      } /* endif */

      fvar++;
   } /* endfor */

   /* Restore the var pointer */
   var = savevar;

   /**************************************************************/
   /*                                                            */
   /* Since some variables are being dropped, the VarChild and   */
   /*  VarLastChild pointers must be fixed up. These pointers    */
   /*  are only used for OCCURS statements for higher level      */
   /*  variables.                                                */
   /*                                                            */
   /**************************************************************/

   for (i=0; i < maxcnv; i++)
   {
      if (var[i].VarChild > 0)
      {
         /* Point to the next variable, which must be the child */
         var[i].VarChild = i + 1;

         /* Start with the first child */
         var[i].VarLastChild = i + 1;

         /* Next, we must find the last child */
         j = i + 2;
         while ((j < maxcnv) && (var[j].VarLevel > var[i].VarLevel))
         {
            /* Still haven't found the last one, update the LastChild pointer */
            var[i].VarLastChild = j;

                        /* Go on to the next child */
                        j++;
         }
      }
   }

   if (DEBUG)
   {
		fprintf(cnvfile,"\nDump of Condensed Variable table\n");

		for (i=0; i < maxcnv; i++)
		{
			/* print out the variable in the listing file */
			fprintf(cnvfile,"%.2d %-32s offset(%.5d) length(%.5d) type(%c) usage(%d)\n",
					i,
					var[i].VarName,
					var[i].VarOffset,
					var[i].VarLen,
					var[i].VarType,
					var[i].VarUsage);

			fprintf(cnvfile,"     next(%d) prev(%d) parent(%d) child(%d) lastchild(%d) redefine(%d) sync(%d) occurs(%d)\n",
					var[i].VarNextLevel,
					var[i].VarPrevLevel,
					var[i].VarParent,
					var[i].VarChild,
					var[i].VarLastChild,
					var[i].VarRedefine,
					var[i].VarSync,
					var[i].VarOccur);
		} /* endfor */
   }

   /* write out the CNV initial stuff */
   fprintf(cnvfile, "         DFHCNV TYPE=ENTRY,RTYPE=PC,RNAME=**PGM**\n");
   fprintf(cnvfile, "         DFHCNV TYPE=SELECT,OPTION=DEFAULT\n");

   /* move through the field definitions table */
   i = 0;
   while (i < maxcnv)
   {
      /* check if this is the last field */
      if (i == maxcnv - 1)
      {
         last = ",LAST=YES";
      }
      else
      {
         last = "";
      } /* endif */

      /* Get the occurs count */
      if (var[i].VarOccur > 0)
      {
         occ = var[i].VarOccur;
      }
      else
      {
         occ = 1;
      }

      /* Calculate the number of varaibles for this OCCURS */
      varcount = var[i].VarLastChild - var[i].VarChild + 1;

      /* Get the next variable number */
      if (var[i].VarChild > 0)
      {
         varnum = var[i].VarChild;
      }
      else
      {
         varnum = i;
      }

      /* Write out the fields */
      writeCNVOccurs(var,varnum, varcount, occ, last, cnvfile);

      /* move on to the next variable */
      if (var[i].VarChild > 0)
      {
         i++;
      }

      i += varcount;
          /* skip the group variable */
   } /* endwhile */
}

/**************************************************************/
/*     Subroutine to build the next complete COBOL statement  */
/*     from one or more source lines.  The statement must     */
/*     start with a level number and end with a period.       */
/*     Comments and continuation lines are handled in this    */
/*     routine as well.  The routine returns either a         */
/*     complete COBOL level statement in one continguous      */
/*     line or a null line, indicating that end of file       */
/*     has been reached.  Leading blanks are ignored, as      */
/*     well as columns 1-7 and 73-80, if present.             */
/**************************************************************/

int buildNextLine(char *line, FILE *commareafile, int *inCounter, int *inFileCount)

{
   int     i;
   int     remaining = MAX_STMT_LEN;
   char    cobolLine[MAX_LINE_LEN];
   char *cline;       /* Current line pointer               */

   /* start with a null statement */
   line[0] = '\0';

   /* loop until either a complete statement is built or  */
   /* end of file is reached.                             */
   while (1)
   {
      if (fgets(cobolLine, MAX_LINE_LEN, commareafile) == NULL)
      {
         /* Indicate end of file has been reached */
         line[0] = '\0';
         return 1;
      }

      /* increment the source line counters */
      (*inCounter)++;
	  (*inFileCount)++;

      /* get rid of the newline character in the line, if present */
      if (cobolLine[strlen(cobolLine) - 1] == '\n')
      {
         cobolLine[strlen(cobolLine) - 1] = '\0';
      }

      /* check that the line is long enough and not a comment */
      if ((strlen(cobolLine) > 7) && (cobolLine[6] != '*') &&
          (cobolLine[6] != '/'))
      {
         /* start at column 8 unless a continuation line */
         if (cobolLine[6] == '-')
         {
            cline = &cobolLine[11];
         }
         else
         {
            cline = &cobolLine[7];
         } /* endif */

         /* skip leading blanks */
         cline = skipBlanks(cline);

         /* Check if the line is blank */
         if (strlen(cline) > 0)
         {
            /* skip trailing blanks */
            i = strlen(cline);
            while ((i > 0) && (cline[i] == ' '))
            {
               cline[i] = '\0';
               i--;
            } /* endwhile */

            if (i > 0)
            {
               if (i < remaining)
               {
                  /* concatenate this to the output line */
                  strcat(line, cline);
                  remaining -= i;
               }
               else
               {
                  /* issue error message - statement too large */
                  /* stop processing - something is wrong */
                  return 1;
               } /* endif */

               /* check if we have found a period at the end */
               i = strlen(line) - 1;
               if (line[i] == '.' )
               {
                  /* remove the period */
                  line[i] = 0;

                  /* return with the statement */
                  break;
               }
               else
               {
                  /* Insert a blank as a delimiter */
                  strcat(line," ");
               } /* endif */
            } /* endif */
         } /* endif */
      } /* endif */
   } /* endwhile */

   return 0;
}

/**************************************************************/
/*                                                            */
/* Routine to write out the contents of a variable table      */
/* entry, primarily for debugging purposes.  A pointer to     */
/* the first entry to be printed and the number of variables  */
/* to be printed are passed as input, as well as a pointer    */
/* to the listing file.                                       */
/*                                                            */
/**************************************************************/

void dumpVar(int varnum, VARSTRUCT *fvar, int count, FILE * msgfile)

{
   int i;

   for (i=varnum; i < varnum + count; i++)
   {
      /* print out the variable in the listing file */
      fprintf(msgfile,"%.2d %-32s offset(%.5d) length(%.5d) type(%c) usage(%d)\n",
              i,
              fvar[i].VarName,
              fvar[i].VarOffset,
              fvar[i].VarLen,
              fvar[i].VarType,
              fvar[i].VarUsage);

      fprintf(msgfile,"     next(%d) prev(%d) parent(%d) child(%d) lastchild(%d) redefine(%d) sync(%d) maxsync(%d) occurs(%d)\n",
              fvar[i].VarNextLevel,
              fvar[i].VarPrevLevel,
              fvar[i].VarParent,
              fvar[i].VarChild,
              fvar[i].VarLastChild,
              fvar[i].VarRedefine,
              fvar[i].VarSync,
              fvar[i].VarMaxSync,
              fvar[i].VarOccur);
   }
}

/**************************************************************/
/*                                                            */
/* Subroutine to finish the processing of any occurs or       */
/*  redefines clauses.  This happens when the level of a      */
/*  variable is less than the previous level, or when end     */
/*  of file has been reached.                                 */
/*                                                            */
/**************************************************************/

void finishOcc(VARSTRUCT *fvar,
                           int occuratlevel[50],
                           int lastlevel,
                           const int level,
                           int *offset,
                           FILE *msgfile)

{
   int  i;                  /* Work variable                      */
   int  tempparent;         /* working variable for sync clauses     */
   int  tempvar;            /* temporary work variable               */
   int  delta;              /* working variable for sync clauses     */
   int  addoffset;          /* working variable for sync clauses     */

    /* Finish any OCCURS and REDEFINES clauses */
    for (i = lastlevel; i >= level; i--)
    {
       if (occuratlevel[i] > 0)
       {
          /* Get the occurs at variable */
          tempvar = occuratlevel[i];

          /* see if we need to add in any slack bytes */
          /* These slack bytes are only needed for    */
          /* group variables which have a variable    */
          /* with a SYNC clause within them, and      */
          /* which contain an OCCURS clause.          */
          /* initialize the number of slack bytes required */
          delta = 0;
          /* check if slack bytes are required */
          if ((fvar[tempvar].VarMaxSync > 1) && (fvar[tempvar].VarOccur > 1))
          {
             /* calculate if slack bytes are needed */
             /*delta = (fvar[tempvar].VarOffset + fvar[tempvar].VarLen) % fvar[tempvar].VarMaxSync;*/
             delta = (fvar[tempvar].VarLen) % fvar[tempvar].VarMaxSync;
             if (delta > 0)
             {
                /* compute the number of slack bytes */
                delta = fvar[tempvar].VarMaxSync - delta;

                /* add the slack bytes to the group length */
                fvar[tempvar].VarLen += delta;

                /* Remember the number of ending slack bytes */
                fvar[tempvar].VarEndSlack = delta;

                /* calculate the total number of slack bytes */
                /*delta = delta * (fvar[tempvar].VarOccur - 1);*/

                /* add the extra slack to the current offset */
                (*offset) += delta;
             }
          }

          /* add the additional occurences to the */
          /* current offset                       */
          addoffset = fvar[tempvar].VarLen * (fvar[tempvar].VarOccur - 1);
          (*offset) += addoffset;

          /* Next, add the occurences to any parents */
          /* Get the first parent                    */
          tempparent = fvar[tempvar].VarParent;

          /* Process all parent variables            */
          while (tempparent > 0)
          {
             /* first add in any added slack bytes   */
             fvar[tempparent].VarLen += delta;

             /* Next, add the additional occurences  */
             fvar[tempparent].VarLen += addoffset;

             /* Get the next parent variable         */
             tempparent = fvar[tempparent].VarParent;
          }

          /* clear the occursatlevel pointer      */
          occuratlevel[i] = 0;
       } /* endif */
    } /* endfor */
}

/**************************************************************/
/*                                                            */
/* Routine to write the input file name to the message file.  */
/*                                                            */
/**************************************************************/

void writeFileName(char *fileName, FILE *msgfile)

{
	/* write the file name to the message file */
	fprintf(msgfile, "Processing file %s\n", fileName);
}

/**************************************************************/
/*                                                            */
/* Routine to write the line counter to the message file.     */
/*                                                            */
/**************************************************************/

void writeLinesRead(int inCount, FILE *msgfile)

{
	/* write the line count to the message file */
   fprintf(msgfile,"Input lines read %5d\n",inCount);
}

/**************************************************************/
/*                                                            */
/* Routine to display the command syntax.                     */
/*                                                            */
/**************************************************************/

void printHelp(char *cmd)
{
   printf("\nBUILDCNV V1.01\n");
   printf("\nCommand usage is:\n\n");
   printf("  %s cobol-file-name <CNV-file-name>\n\n", cmd);
   printf("The file name parameter can include multiple file names, each\n");
   printf("one separated by a plus (+) character.\n");
   printf("The CNV file name parameter is the name of the new file to write\n");
   printf("the DFHCNV macros into.  If not specified, then the\n");
   printf("first cobol file name is used, with an extension of CNV.\n");
}

/**************************************************************/
/*                                                            */
/*                                                            */
/*                                                            */
/**************************************************************/

int main(int argc, char *argv[])

{
   int  offset=0;           /* Offset within the commarea         */
   int  length=0;           /* Length of data item                */
   int  level;              /* Level number of current line       */
   int  redefLevel=0;       /* Level number of redefines clause   */
   int  maxvar=0;           /* Number of fields found             */
   int  i;                  /* Work variable                      */
   char use;                /* Type of USAGE                      */
   char *cline;             /* Current line pointer               */
   char *endLevel;          /* Pointer to end of level number     */
   char *varname;           /* Pointer to variable name           */
   char *redefines;         /* Location of REDEFINES              */
   char *pic;               /* Location of PIC                    */
   char *occurs;            /* Location of OCCURS                 */
   char *sync;              /* Location of SYNC                   */
   char *just;              /* Location of JUST                   */
   char *comp;              /* Location of COMP                   */
   char *binary;            /* Location of BINARY                 */
   char *packed;            /* Location of PACKED                 */
   char *display;           /* Location of DISPLAY                */
   char *index;             /* Location of INDEX                  */
   char *pointer;           /* Location of POINTER                */
   char *procptr;           /* Location of PROCEDURE-POINTER      */
   char *usage;             /* Location of USAGE                  */
   char *blank;             /* Location of BLANK                  */
   char *value;             /* Location of VALUE                  */
   char *fileExt;           /* file extension                     */
   char *nextFile;			/* pointer to next input file name    */
   char *nextTemp;          /* temporary pointer to next input file */
   char datatype;           /* type of data field                 */
   char redefineName[MAX_NAME_LEN + 1]; /* redefines target       */
   char commFileName[4096]; /* name of COMMAREA copy book file    */
   char msgFileName[255];   /* name of message file               */
   char cnvFileName[255];   /* name of DFHCNV translation file       */
   char cobolStmt[MAX_STMT_LEN + 1];  /* current commarea line       */
   VARSTRUCT *fvar;         /* variable table pointer                */
   int  firstatlevel[50];   /* keep track of first variable at level */
   int  lastatlevel[50];    /* keep track of last variable at level  */
   int  occuratlevel[50];   /* keep track of OCCURS at level         */
   int  savelast;           /* working variable                      */
   int  redefvar;           /* working variable                      */
   int  saveParent=0;       /* remember variable's parent            */
   int  lastlevel=0;        /* remember level of previous variable   */
   int  tempvar;            /* temporary work variable               */
   int  align;              /* working variable for sync clauses     */
   int  delta;              /* working variable for sync clauses     */
   int  tempchild;          /* working variable for sync clauses     */
   int  eof;                /* End of File indicator                 */
   int	RC=0;               /* Return code from program              */

 /*
  * File variables and counters
  */

    FILE        *commareafile; /* input COBOL program         */
    FILE        *msgfile;      /* conversion program messages */
    FILE        *cnvfile;      /* conversion table file       */
    int         inCounter=0;   /* input line counter          */
    int         outCounter=0;  /* output line counter         */
	int			inFile=0;      /* Lines within current file   */

/**************************************************************/
/*     Allocate and initialize variable areas                 */
/**************************************************************/

    fvar = (VARSTRUCT *)malloc(sizeof(VARSTRUCT) * (MAX_VAR_COUNT + 1));
    memset(fvar,'\0',(sizeof(VARSTRUCT) * (MAX_VAR_COUNT + 1)));
    for (i=0; i < 50; i++)
    {
       firstatlevel[i] = 0;
       lastatlevel[i]  = 0;
       occuratlevel[i] = 0;
    } /* endfor */

/**************************************************************/
/*     Parse the file name and generate new file name.        */
/**************************************************************/

   if (argc >= 2)        /* check for file name */
   {
	  /* Get the file name parameter, making sure not to overflow */
      strncpy(commFileName,argv[1],sizeof(commFileName));

	  /* Next, we must check for multiple file names */
	  /* Check for the delimiter character           */
	  nextFile = strchr(commFileName,'+');
	  if (nextFile != NULL)
	  {
		  /* Found a delimiter, terminate the first file name */
		  nextFile[0] = '\0';

		  /* Point to the rest of the file name parameter */
		  nextFile++;
	  }

      /* check for help request */
      if ((commFileName[0] == '?') || (commFileName[1] == '?'))
      {
         printHelp(argv[0]);
         exit(1);
      } /* endif */
   }
   else
   {
      printHelp(argv[0]);
      exit(1);
   } /* endif */


/**************************************************************/
/* Generate the new file name                                 */
/**************************************************************/

   /* get the full file name */
   strcpy(cnvFileName,commFileName);

   /* look for a period */
   fileExt = strchr(cnvFileName, '.');
   if (fileExt != NULL)
   {
      /* remove the file extension if it exists */
      (*fileExt) = 0;
   } /* endif */

   /* get the file name for the message and list files */
   strcpy(msgFileName, cnvFileName);

   /* add a file extension to the new DFHCNV and message files */
   strcat(cnvFileName,".CNV");
   strcat(msgFileName,".msg");

   /* Check for a second parameter */
   if (argc >= 3)        /* check for file name */
   {
	   /* Get the name of the new DFHCNV macro file */
	   strncpy(cnvFileName, argv[2], sizeof(cnvFileName));
	   strncpy(msgFileName, argv[2], sizeof(cnvFileName));

	   /* Remove any extension from the message file name */
	   /* Check for the extension */
	   fileExt = strchr(msgFileName, '.');
	   /* Did we find one? */
	   if (fileExt != NULL)
	   {
		   /* Remove the file extension */
		   (*fileExt) = 0;

		   /* Append a file extension */
		   strcat(msgFileName,".msg");
	   }
   }

/**************************************************************/
/* Open the cobol program, new file and message file.         */
/**************************************************************/

   if ((commareafile = fopen(commFileName, "r")) == NULL)
   {
	  printf("Unable to open input file - %s\n", commFileName);
      exit(100);
   }

   if ((msgfile = fopen(msgFileName, "w")) == NULL)
   {
      fclose(commareafile);
      exit(100);
   }

   if ((cnvfile = fopen(cnvFileName, "w")) == NULL)
   {
      fclose(msgfile);
      fclose(commareafile);
      exit(100);
   }

/* Write the name of the input file to the message file */
	writeFileName(commFileName, msgfile);

/**************************************************************/
/*                                                            */
/*  This routine will read the COMMAREA copy book file        */
/*  and generate a listing file and CNV macros to build       */
/*  the CICS conversion template.                             */
/*                                                            */
/**************************************************************/

   /* Read the next line(s) in the source file. This routine */
   /* returns a complete COBOL source statement, with the    */
   /* ending period removed.  It also removes comments and   */
   /* and blank lines.  It returns a 1 if end of file is     */
   /* reached.  Leading and trailing blanks are removed.     */
   while (1)
   {
	  /* Get the next COBOL line */
      eof = buildNextLine(cobolStmt, commareafile, &inCounter, &inFile);

	  /* Check for end of file */
	  if (eof == 1)
	  {
		  /* First, close the current file */
		  fclose(commareafile);

		  /* Write the number of lines read to the message file */
		  writeLinesRead(inFile, msgfile);

		  /* Reset the file line counter */
		  inFile = 0;

		  /* Check if more file names were specified in the parameter */
		  if (nextFile == NULL)
		  {
			  /* No more files, all done */
			  break;
		  }

		  /* Check for any more delimiters */
		  nextTemp = strchr(nextFile, '+');

		  /* Did we find any? */
		  if (nextTemp != NULL)
		  {
			  /* Delimit the next file name */
			  nextTemp[0] = '\0';

			  /* Point to the next file name in the parameter */
			  nextTemp++;
		  }

		  /* Move the next file name into the file name area */
		  strcpy(commFileName, nextFile);

		  if ((commareafile = fopen(commFileName, "r")) == NULL)
		  {
			  /* Can't open the file, tell the user */
			 printf("Unable to open input file - %s\n", commFileName);
			 fprintf(msgfile,"Unable to open input file - %s\n", commFileName);

			 /* Set a file not found return code */
			 RC = 100;

			 /* No more work to do - exit gracefully */
			 break;
		  }

		  /* Write the name of the input file to the message file */
		  writeFileName(commFileName, msgfile);

		  /* Point to the next file name in the parameter */
		  nextFile = nextTemp;
	  }

      /* find the end of the level number       */
      endLevel = &cobolStmt[1];
      while ((endLevel[0] != ' ') && (endLevel[0] != '\0') && (eof == 0))
      {
         endLevel++;
      } /* endwhile */

      if ((strlen(endLevel) > 0) && (eof == 0))
      {
         /* terminate the level number string */
         endLevel[0] = '\0';

         /* convert the number to a level number */
         level = atoi(cobolStmt);

         /* point to the rest of the line */
         cline = endLevel + 1;

         /* point to variable name */
         cline = skipBlanks(cline);

         /* search for various keywords */
         pic       = strstr(cline, PIC_STR);
         occurs    = strstr(cline, OCCURS_STR);
         sync      = strstr(cline, SYNC_STR);
         just      = strstr(cline, JUST_STR);
         comp      = strstr(cline, COMP_STR);
         binary    = strstr(cline, BINARY_STR);
         packed    = strstr(cline, PACKED_STR);
         display   = strstr(cline, DISPLAY_STR);
         index     = strstr(cline, INDEX_STR);
         pointer   = strstr(cline, POINTER_STR);
         procptr   = strstr(cline, PROCEDURE_STR);
         redefines = strstr(cline, REDEFINES_STR);
         usage     = strstr(cline, USAGE_STR);
         blank     = strstr(cline, BLANK_STR);
         value     = strstr(cline, VALUE_STR);

         /* find the end of the variable name */
         varname = cline;
         while ((cline[0] != ' ') && (cline[0] != '\0') &&
                (cline[0] != '.'))
         {
            cline++;
         } /* endwhile */

         /* Check for a COBOL keyword instead of a variable name */
         /* In COBOL, it is legitimate to leave off the variable */
         /* name and let it default to FILLER.  The purpose of   */
         /* this subroutine is to recognize if we have a COBOL   */
         /* keyword immediately following the level, which       */
         /* indicates that the variable name is defaulting to    */
         /* FILLER.  If this is the case, then the pointer is    */
         /* changed, and we reset the end of string              */
         if (strcmp(cline,REDEFINES_STR) == 0)
         {
            varname = FILLER_STR;
         } /* endif */
         else
         {
            /* Terminate the variable name to produce a string */
            if (cline[0] != '\0')
            {
               cline[0] = '\0';
               cline++;
            }
         }

         if (redefLevel >= level)
         {
            redefLevel = 0;
         }

         /* Ignore if not a level 01-49 */
         if ((level > 0) && (level <= 49))
         {
            /* count the number of fields found */
            maxvar++;

            if (redefines != NULL)
            {
               /* get the variable name in the redefines clause */
               processRedefine(redefines, redefineName);

               /* check if it is the last at the same level */
               redefvar = lastatlevel[level];
               while ((strcmp(fvar[redefvar].VarName, redefineName) != 0) &&
                      (redefvar > 0))
               {
                  /* Not last at same level - try to find it anyway */
                  redefvar = fvar[redefvar].VarPrevLevel;
               } /* endif */

               if (redefvar == 0)
               {
                  /* write error to message file - invalid redefines */
                  fprintf(msgfile, "Invalid redefines - %d %s redefines %s",
                          level, varname, redefineName);
               } /* endif */

               if ((level < redefLevel) || (redefLevel == 0))
               {
                  redefLevel = level;
               }

               /* update the redefines pointer if necessary */
               fvar[maxvar].VarRedefine = redefvar;
            } /* endif */

            /* Insert into the level pointers */
            if (firstatlevel[level] == 0)
            {
               /* first at level, create chain */
               firstatlevel[level] = maxvar;
               lastatlevel[level]  = maxvar;
            }
            else
            {
               /* update the backwards chain pointers */
               savelast = lastatlevel[level];
               lastatlevel[level] = maxvar;
               fvar[maxvar].VarPrevLevel = savelast;

               /* update the forward level chain pointers */
               fvar[savelast].VarNextLevel = maxvar;
               fvar[maxvar].VarNextLevel = 0;
            } /* endif */

            /* did the level increase or decrease? */
            if (level > lastlevel)
            {
               /* level is greater than previous variable */
               /* This indicates that we have moved down  */
               /* in the hierarchy.  Occurs and redefines */
               /* will stay the same, and the previous    */
               /* variable is the new parent.             */
               saveParent = maxvar - 1;

               /* Indicate that this variable is a parent */
               if (fvar[saveParent].VarChild == 0)
               {
                  fvar[saveParent].VarChild = maxvar;
               }

               /* Remember the parent variable            */
               fvar[maxvar].VarParent = saveParent;

               /* Set the current level                   */
               lastlevel = level;
            }
            else
            {
               /* Handle less than or same level.        */
               /* Level is less than previous variable   */
               /* This indicates that we have moved up   */
               /* the hierarchy.  This may indicate that */
               /* one or more redefines and/or occurs    */
               /* clauses are now finished.  For the     */
               /* occurs clauses, we need to calculate   */
               /* a new offset, respecting the alignment */
               /* of the occurs clause.  If a redefines  */
               /* clause has ended, the lengths of the   */
               /* two areas must be checked, and the     */
               /* offset adjusted if necessary.  Lastly, */
               /* we must find the new parent, if any.   */

               /* Finish any OCCURS and REDEFINES clauses */
               finishOcc(fvar, &occuratlevel[0], lastlevel, level, &offset, msgfile);

               /* Find the new parent and fill in the LastChild fields */
               while ((level < lastlevel) && (saveParent > 0))
               {
                  fvar[saveParent].VarLastChild = maxvar - 1;
                  lastlevel = fvar[saveParent].VarLevel;
                  saveParent = fvar[saveParent].VarParent;
               } /* endwhile */
            } /* endif */

            /* Reset the offset for any REDEFINES clauses */
            if ((redefines != NULL) && (redefvar > 0))
            {
               /* Reset the offset for a new redefines */
               offset = fvar[redefvar].VarOffset;
            } /* endif */

            /* insert the field into the variable table */
            strcpy(fvar[maxvar].VarName,varname);
            fvar[maxvar].VarLevel  = level;
            fvar[maxvar].VarOffset = offset;

            /* save the current variable's parent */
            fvar[maxvar].VarParent = saveParent;

            /* Check for OCCURS clause   */
            if (occurs != NULL)
            {
               /* Fill in the occurs count */
               fvar[maxvar].VarOccur = handleOccurs(occurs);
               if (fvar[maxvar].VarOccur > 1)
               {
                  /* remember the occurs at this level */
                  occuratlevel[level] = maxvar;
               }
               else
               {
                  fvar[maxvar].VarOccur = 0;
               }
            } /* endif */

            /* handle the picture clause */
            if (pic != NULL)
            {
               length = handlePic(pic, &datatype);
            }
            else
            {
               length = 0;
            }

            /* handle the usage clause keywords */
            /* set the default to no usage type */
            use = 0;

            /* check for USAGE DISPLAY */
            if (display != NULL)
            {
               /* set the usage */
               use = USE_DISPLAY;
            }

            /* check for USAGE BINARY */
            if (binary != NULL)
            {
               /* set the usage */
               use = USE_BINARY;

               /* calculate the actual length of the item */
               length = handleBinaryLength(length);
            }

            /* check for USAGE PACKED */
            if (packed != NULL)
            {
               /* set the usage */
               use = USE_PACKED;

               /* calculate the actual length of the item */
               length = handlePackedLength(length);
            }

            /* check for one of the USAGE COMP keywords */
            if (comp != NULL)
            {
               use = handleComp(comp, &length);
            }

            /* check for USAGE INDEX */
            if (index != NULL)
            {
               /* set the usage and the length */
               length = 4;
               use = USE_INDEX;
            }

            /* check for USAGE POINTER or PROCEDURE-POINTER */
            if ((pointer != NULL) || (procptr != NULL))
            {
               /* set the usage and the length */
               length = 4;
               use = USE_POINTER;
            }

            /* Set the fields in the variable record */
            fvar[maxvar].VarLen    = length;
            fvar[maxvar].VarType   = datatype;
            fvar[maxvar].VarUsage  = use;

            /* calculate the alignment of this variable, in */
            /* case there is a sync clause either for this  */
            /* variable or for a parent of this variable    */

            /* get the boundary to align to */
            switch (use)
            {
               case USE_POINTER:
               case USE_INDEX:
               case USE_BINARY:
               case USE_FLOAT1:
               case USE_FLOAT2:
                  {
                     /* align on 2, 4 or 8 byte boundary */
                     align = length;
                     break;
                  }

               default:
                  {
                     /* align on single byte boundary */
                     align = 1;
                     break;
                  }
            }

            /* Check for a SYNC or SYNCHRONIZED clause    */
            /* If none was found for this variable, check */
            /* for one at a higher level                  */

            if (sync != NULL)
            {
               /* save the alignment, if required */
               if (align > 1)
               {
                  fvar[maxvar].VarSync = align;

                  /* Check if we need to propagate the maximum */
                  /*  requested alignment to any parents.      */
                  /* We do not need to worry about level 2     */
                  /* or level 1 variables                      */

                  /* Get the parent */
                  tempvar = fvar[maxvar].VarParent;

                  /* Check that the level is greater than 2 and */
                  /*  that the variable has parents             */
                  if ((level > 2) && (tempvar > 0))
                  {
                     /* Ensure we align any parents with occurs */
                     /* clauses correctly                       */
                     while (tempvar > 0)
                     {
                        fvar[tempvar].VarSync = align;

                        /* check if this is the maximum alignment */
                        /* for this variable                      */
                        if (fvar[tempvar].VarMaxSync < align)
                        {
                           /* Remember the maximum alignment within  */
                           /* this group variable, in case we need   */
                           /* to handle slack bytes due to an OCCURS */
                           /* clause                                 */
                           fvar[tempvar].VarMaxSync = align;
                        }
                        /* move on to the next parent */
                        tempvar = fvar[tempvar].VarParent;
                     }
                  }
               }
            }
            else
            {
               /* check for any SYNC clauses in parents */
               /* First, check if this variable has a parent */
               if (fvar[maxvar].VarParent > 0)
               {
                  /* Get the parent */
                  tempvar = fvar[maxvar].VarParent;

                  /* Check if this is the first child and */
                  /*  if the parent has a SYNC clause     */
                  if ((fvar[tempvar].VarChild == maxvar) && (fvar[tempvar].VarSync > 0))
                  {
                     /* Yes to both, propagate the sync clause down */
                     /* see which one has the greater alignment     */
                     if (align < fvar[tempvar].VarSync)
                     {
                        align = fvar[tempvar].VarSync;
                     }

                     fvar[maxvar].VarSync = align;
                  }
               }
            }

            /* Does this variable need to be aligned? */
            /* if so, then align the variable and     */
            /* then propagate the alignment to any    */
            /* parents if necessary.  Alignment of    */
            /* parents is only necessary if this is   */
            /* the first child.                       */
            /* These slack bytes are added before the */
            /* variable.  Other slack bytes may be    */
            /* added after a group variable with an   */
            /* OCCURS clause.  These will be added in */
            /* the routine above which handles        */
            /* changes in level.                      */
            if (fvar[maxvar].VarSync > 0)
            {
               /* change the offset to align the variable, */
               /*  if not already aligned                  */
               if ((fvar[maxvar].VarOffset%align) > 0)
               {
                  /* calculate the number of slack bytes required */
                  delta = align - (fvar[maxvar].VarOffset%align);

                  /* change the offset to align the variable */
                  fvar[maxvar].VarOffset  += delta;

                  /* increase the current offset to match    */
                  offset += delta;

                  /* Next, if this is the first child, propagate */
                  /* the alignment up to the parent              */
                  tempvar = fvar[maxvar].VarParent;
                  tempchild = maxvar;
                  while (tempvar > 0)
                  {
                     if (fvar[tempvar].VarChild == tempchild)
                     {
                        /* remember the alignment level */
                        fvar[tempvar].VarSync = align;

                        /* change the offset to align the variable */
                        fvar[tempvar].VarOffset = fvar[tempchild].VarOffset;
                     }
                     else
                     {
                        fvar[tempvar].VarLen += delta;
                     }

                     /* move up to the next higher variable and see */
                     /* if we are still the first child.            */
                     tempchild = tempvar;
                     tempvar = fvar[tempchild].VarParent;
                  }
               }
            }

            /* Add the length to the lengths of any parents */
			if (DEBUG)
			{
				fprintf(msgfile,"adding length to parents for variable %d\n",maxvar);
				dumpVar(maxvar,fvar,1,msgfile);
			}

            i = saveParent;
            while (i > 0)
            {
				if (DEBUG)
				{
					fprintf(msgfile,"parent before ");
					dumpVar(i,fvar,1,msgfile);
				}

                fvar[i].VarLen += length;

				if (DEBUG)
				{
					fprintf(msgfile,"        after ");
					dumpVar(i,fvar,1,msgfile);
				}

               i = fvar[i].VarParent;
            } /* endwhile */


            /* print out the variable in the listing file if debug enabled */
			if (DEBUG)
			{
				fprintf(msgfile,"%.2d %-32s offset(%.5d) length(%.5d) type(%c) usage(%d)\n",
					    level, varname, offset, length, datatype, use);

				fprintf(msgfile,"     var(%d) next(%d) prev(%d) parent(%d) child(%d) redefine(%d) sync(%d) maxsync(%d) occurs(%d) \n",
					    maxvar,
						fvar[maxvar].VarNextLevel,
						fvar[maxvar].VarPrevLevel,
						fvar[maxvar].VarParent,
						fvar[maxvar].VarChild,
						fvar[maxvar].VarRedefine,
						fvar[maxvar].VarSync,
						fvar[maxvar].VarMaxSync,
						fvar[maxvar].VarOccur);
			}

            /* increment the offset by the length of the variable */
            offset += length;
         }

	     /* increment the line counter for the output program */
	     outCounter++;
      } /* endif */

   } /* endwhile */

   /* Finish up any unfinished OCCURS and REDEFINES clauses */
   finishOcc(fvar, &occuratlevel[0], lastlevel, 1, &offset, msgfile);

   if (DEBUG)
   {
		/* print all the variables, starting with the first variable */
		fprintf(msgfile,"\nDump of Variable table\n");
		dumpVar(1, fvar, maxvar, msgfile);

	    fprintf(msgfile,"\nFirst at Level table\n");
		for (i=1; i < 50; i++)
		{
			if (firstatlevel[i] > 0)
			{
				fprintf(msgfile,"  level %.2d - first(%d)\n",i,firstatlevel[i]);
			}
		}

		fprintf(msgfile,"\nLast at Level table\n");
		for (i=1; i < 50; i++)
		{
			if (lastatlevel[i] > 0)
			{
				fprintf(msgfile,"  level %.2d -  last(%d)\n",i,lastatlevel[i]);
			}
		}

		fprintf(msgfile,"\nOccurs table\n");
		for (i=1; i < 50; i++)
		{
			if (occuratlevel[i] > 0)
			{
				fprintf(msgfile,"  level %.2d - occurs(%d)\n",i,occuratlevel[i]);
			}
		}
   }

   /* Build the DFHCNV translation file for this commarea */
   buildCNV(cnvfile, fvar, maxvar);

   /* add the total length of the record to the list file */
   fprintf(msgfile,"\nTotal length of area in bytes %d\n", offset);

   /* write the counters to the message file */
   fprintf(msgfile,"\nTotals\nInput lines read %5d\n",inCounter);
   fprintf(msgfile,"Output lines written %5d\n",outCounter);

   /* close the files */
   fclose(cnvfile);
   fclose(msgfile);

/**************************************************************/
/* Finished - return normally                                 */
/**************************************************************/

   return RC;
}

/*-------------*/
/* End Of File */
/*-------------*/
